# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import base64
import hashlib
import http
import json
import logging
import os
import re
import socket
import time
import webbrowser
from contextlib import closing
from functools import partial
from http.server import BaseHTTPRequestHandler
from http.server import HTTPServer
from typing import Optional
from urllib.parse import parse_qs
from urllib.parse import urlparse

from requests import HTTPError
from requests import post
from requests.auth import HTTPBasicAuth
from requests_oauthlib import OAuth2Session
from vdk.plugin.control_api_auth.auth_exception import VDKLoginFailedError
from vdk.plugin.control_api_auth.auth_request_values import AuthRequestValues
from vdk.plugin.control_api_auth.base_auth import BaseAuth
from vdk.plugin.control_api_auth.login_types import LoginTypes

log = logging.getLogger(__name__)


def generate_pkce_codes():
    """
    Generate PKCE code challenge and code verifier necessary during Authorization Code Workflow
    as described in RFC 7636 (see  https://tools.ietf.org/html/rfc7636)
    :return: code_verifier, code_challenge, code_challenge_method
    """
    code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode("utf-8")
    code_verifier = re.sub("[^a-zA-Z0-9]+", "", code_verifier)
    code_challenge = hashlib.sha256(code_verifier.encode("utf-8")).digest()
    code_challenge = base64.urlsafe_b64encode(code_challenge).decode("utf-8")
    code_challenge = code_challenge.replace("=", "")
    return code_verifier, code_challenge, "S256"


class LoginHandler:
    """
    Class to handle the updates of the authentication parameters and the log
    in process itself.
    """

    CODE_PARAMETER_KEY = "code"
    STATE_PARAMETER_KEY = "state"
    REFRESH_TOKEN_KEY = "refresh_token"  # nosec
    GRANT_TYPE = "authorization_code"

    def __init__(
        self,
        client_id,
        client_secret,
        exchange_endpoint,
        redirect_uri,
        code_verifier,
        auth,
    ):
        """
        :param client_id:
            The client identifier;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-2.3.1
        :param client_secret:
            The client identifier;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-2.3.1
        :param exchange_endpoint:
            The URL which exchanges token for access token.
        :param redirect_uri:
            URL for the authorization server to redirect back to (as defined
            in https://datatracker.ietf.org/doc/html/rfc6749#section-3.1.2).
        :param code_verifier:
            A string used to correlate the authorization request to the token
            request as described in https://tools.ietf.org/html/rfc7636
        :param auth:
            A BaseAuth object.
        """
        self.client_id: str = client_id
        self.client_secret: str = client_secret
        self.exchange_endpoint: str = exchange_endpoint
        self.redirect_uri: str = redirect_uri
        self.code_verifier: str = code_verifier
        self.auth: BaseAuth = auth
        self.login_exception: Optional[Exception] = None

    def login_with_authorization_code(self, path):
        try:
            auth_code = self._acquire_auth_code(path)
            json_content = self._exchange_code_for_tokens(auth_code)
            self.auth.update_oauth2_authorization_url(self.exchange_endpoint)
            if self.REFRESH_TOKEN_KEY in json_content:
                self.auth.update_refresh_token(json_content[self.REFRESH_TOKEN_KEY])
            self.auth.update_access_token(
                json_content[AuthRequestValues.ACCESS_TOKEN_KEY.value]
            )
            self.auth.update_access_token_expiration_time(
                time.time()
                + int(json_content[AuthRequestValues.EXPIRATION_TIME_KEY.value])
            )
            self.auth.update_client_id(self.client_id)
            self.auth.update_client_secret(self.client_secret)
            self.auth.update_auth_type(LoginTypes.CREDENTIALS.value)
        except Exception as login_exception:
            self.login_exception = login_exception
            raise

    def _exchange_code_for_tokens(self, auth_code):
        headers = {
            AuthRequestValues.CONTENT_TYPE_HEADER.value: AuthRequestValues.CONTENT_TYPE_URLENCODED.value,
        }
        data = (
            f"code={auth_code}&"
            + f"grant_type={self.GRANT_TYPE}&"
            + f"redirect_uri={self.redirect_uri}"
        )
        if not self.client_secret:
            log.debug(
                "No client secret specified. We assume native app workflow with PKCE (RFC 7636)."
            )
            data = f"{data}&code_verifier={self.code_verifier}"
        basic_auth = HTTPBasicAuth(self.client_id, self.client_secret)
        try:
            response = post(
                self.exchange_endpoint, data=data, headers=headers, auth=basic_auth
            )
            response.raise_for_status()
            json_data = json.loads(response.text)
        except HTTPError as http_exception:
            raise VDKLoginFailedError(
                what="Failed to login.",
                why="HTTP error occurred during authorization workflow. "
                f"Error was: HTTP error {http_exception.response.status_code}: {http_exception.response.content}",
                consequence="Operations may not work unless previous login is still valid.",
                countermeasure="Try to login again.\n"
                "  If problem persist, try to see the reason in the why section and instruction there.\n"
                "  If that does not help, open ticket to the support team. "
                "Provide all logs you have and describe exact steps to reproduce the issue "
                "and commands executed.",
            )
        except Exception as e:
            raise VDKLoginFailedError(
                what=f"Failed to login: {str(e)}.",
                why="Problem in the configuration or service. Cannot acquire tokens.",
                consequence="Cannot login user.",
                countermeasure="Contact the owner to resolve the problem.",
            )
        return json_data

    def _acquire_auth_code(self, path):
        url = urlparse(path)
        query_components = parse_qs(url.query)
        auth_code = ""
        state = ""
        if self.CODE_PARAMETER_KEY in query_components:
            auth_code = query_components[self.CODE_PARAMETER_KEY][0]
        if self.STATE_PARAMETER_KEY in query_components:
            state = query_components[self.STATE_PARAMETER_KEY][0]
        if state != AuthRequestValues.STATE_PARAMETER_VALUE.value or not state:
            raise VDKLoginFailedError(
                what="Failed to login.",
                why="Possibly the request was intercepted.",
                consequence="Cannot login user.",
                countermeasure="Try to login again.",
            )
        if not auth_code:
            raise VDKLoginFailedError(
                what="Authentication code is empty",
                why="The user failed to authenticate properly.",
                consequence="User will not be logged in.",
                countermeasure="Try to login again.",
            )
        return auth_code


class MyHttpRequestHandler(BaseHTTPRequestHandler):
    """
    Class used by RedirectAuthentication to handle the GET redirect request in order to acquire the refresh and access
    tokens. In essence the class creates Authentication object which fills the necessary configuration fields for the
    credentials authentication type.
    """

    UTF_ENCODING = "utf8"
    CONTENT_TYPE_TEXT_HTML = "text/html"
    HTML_LOGIN_SUCCESS_TEMPLATE = (
        "<html><head></head><body><h1>Login Successful!</h1></body></html>"
    )
    HTML_LOGIN_FAILURE_TEMPLATE = "<html><head></head><body><h1>Login Failed. Check terminal for more information.</h1></body></html>"

    def __init__(self, login_handler, *args, **kwargs):
        self.login_handler = login_handler
        super().__init__(*args, **kwargs)

    def do_GET(self):
        self.send_response(http.HTTPStatus.OK)
        self.send_header(
            AuthRequestValues.CONTENT_TYPE_HEADER.value, self.CONTENT_TYPE_TEXT_HTML
        )
        self.end_headers()
        try:
            self.login_handler.login_with_authorization_code(self.path)
            self.wfile.write(bytes(self.HTML_LOGIN_SUCCESS_TEMPLATE, self.UTF_ENCODING))
        except:
            self.wfile.write(bytes(self.HTML_LOGIN_FAILURE_TEMPLATE, self.UTF_ENCODING))
            raise


class RedirectAuthentication:
    """
    Class used to start web browser and http process which will handle single GET request and shut down.
    The browser will redirect to the http process which handles the request.

    Making the port which we use for authentication constant enable integration with some OAuth2 authorization server
    (e.g VMware Cloud) which require a redirect uri when creating OAuth2 app which needs to be set to the exact
    value which is used in the redirect url. Before we took random port from the ones available which prevented us
    from configuring the URI in the application options.
    It might be good idea for this to be configurable, so we can switch to random ports when necessary
    """

    def __init__(
        self,
        client_id,
        client_secret,
        oauth2_discovery_url,
        oauth2_exchange_url,
        auth,
        redirect_uri="http://127.0.0.1",
        redirect_uri_default_port=31113,
    ):
        """

        :param client_id:
            The client identifier of the OAuth2 Application. Find out more https://tools.ietf.org/html/rfc6749#section-2.3.1
        :param client_secret:
            The client secret of the OAuth2 Application. Find out more https://tools.ietf.org/html/rfc6749#section-2.3.1
        :param oauth2_discovery_url:
            Token or Authorization URI used to exchange grant for access token
        :param oauth2_exchange_url:
            The URL which exchanges token for access token.
        :param auth:
            A BaseAuth object.
        :param redirect_uri:
            The redirect uri which will be used in Authorization Workflow.
            Per https://tools.ietf.org/html/rfc8252#section-7.3 it should be http://127.0.0.1
            so there should not be reason to override the default except for tests
        :param redirect_uri_default_port:
            The default port to use for redirect uri unless env. variable 'OAUTH2_REDIRECT_URI_PORT' is used.
            If None - then random one is assigned
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.oauth2_discovery_url = oauth2_discovery_url
        self.oauth2_exchange_url = oauth2_exchange_url
        self.auth = auth

        env_port = os.getenv("OAUTH2_REDIRECT_URI_PORT", default=None)
        if env_port:
            self.port = int(env_port)
        elif redirect_uri_default_port is not None:
            self.port = redirect_uri_default_port
        else:
            self.port = self.find_free_port()
        self.redirect_uri = f"{redirect_uri}:{self.port}"

        (
            self.code_verifier,
            self.code_challenge,
            self.code_challenge_method,
        ) = generate_pkce_codes()

    def authentication_process(self):
        authorization_url = self._create_authorization_redirect_url()
        discovery_endpoint = authorization_url[0]
        login_handler = self._create_login_handler()
        handler = self._create_redirect_handler(login_handler)
        self._redirect(discovery_endpoint, handler, login_handler)

    def _redirect(self, discovery_endpoint: str, handler, login_handler: LoginHandler):
        with HTTPServer(("", self.port), handler) as server:
            print(f"Opening browser at:\n{discovery_endpoint}")
            is_open = webbrowser.open(discovery_endpoint)
            if not is_open:
                print(
                    "We failed to open the browser automatically and will proceed to login manually.\n"
                    "Please, follow below instructions:"
                )
                self._manual_login(discovery_endpoint, login_handler)
            else:  # TODO: that's not very good UX, let's timeout after 1 minute
                print(
                    "Press [Ctrl + C]/[Command + C] to quit in case of error in the browser."
                )
                server.handle_request()
                if login_handler.login_exception:
                    raise login_handler.login_exception

    @staticmethod
    def _create_redirect_handler(login_handler: LoginHandler):
        return partial(MyHttpRequestHandler, login_handler)

    def _create_login_handler(self) -> LoginHandler:
        return LoginHandler(
            client_id=self.client_id,
            client_secret=self.client_secret,
            exchange_endpoint=self.oauth2_exchange_url,
            redirect_uri=self.redirect_uri,
            code_verifier=self.code_verifier,
            auth=self.auth,
        )

    def _create_authorization_redirect_url(self):
        oauth = OAuth2Session(client_id=self.client_id, redirect_uri=self.redirect_uri)
        if not self.client_secret:
            log.debug(
                "No client secret specified. We assume native app workflow with PKCE (RFC 7636)."
            )
            return oauth.authorization_url(
                self.oauth2_discovery_url,
                state=AuthRequestValues.STATE_PARAMETER_VALUE.value,
                prompt=AuthRequestValues.LOGIN_PROMPT.value,
                code_challenge=self.code_challenge,
                code_challenge_method=self.code_challenge_method,
            )
        else:
            return oauth.authorization_url(
                self.oauth2_discovery_url,
                state=AuthRequestValues.STATE_PARAMETER_VALUE.value,
                prompt=AuthRequestValues.LOGIN_PROMPT.value,
            )

    @staticmethod
    def find_free_port():
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]

    def _manual_login(self, discovery_endpoint, handler: LoginHandler):
        # TODO: Replace the print statements with something more user friendly.
        # manual login is necessary for environment where there is not a browser -
        # console only OS environment, WSL (Windows subsystem for linux)
        print(
            f"Copy paste the following link in your browser:\n\n{discovery_endpoint}\n\n"
        )
        print("Login using your company credentials and wait all redirects to finish.")
        print(
            f"The last redirect will be to a page that starts with {self.redirect_uri} - "
            "the page may show an error that site cannot be reached which you can ignore.\n"
        )
        print(
            f"Please, copy the address that the browser was redirected to (it should start with {self.redirect_uri}) and paste it here:"
        )
        url = str(input("Copy-pasted URL: ")).strip()
        handler.login_with_authorization_code(url)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from enum import Enum


class AuthRequestValues(Enum):
    """
    An enum used to the common values used in the redirect logic.
    """

    STATE_PARAMETER_VALUE = "requested"
    LOGIN_PROMPT = "login"
    CONTENT_TYPE_URLENCODED = "application/x-www-form-urlencoded"
    REFRESH_TOKEN_GRANT_TYPE = "refresh_token"  # nosec
    CONTENT_TYPE_HEADER = "Content-Type"
    EXPIRATION_TIME_KEY = "expires_in"
    ACCESS_TOKEN_KEY = "access_token"  # nosec



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from enum import Enum


class LoginTypes(Enum):
    """
    An enum used to store the types of authentication which the
    versatile data kit sdk (vdk) provides.
    """

    CREDENTIALS = "credentials"
    API_TOKEN = "api-token"  # nosec



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import time
from typing import Optional
from typing import Union

from requests import post
from requests.auth import HTTPBasicAuth
from requests_oauthlib import OAuth2Session
from vdk.plugin.control_api_auth.auth_config import AuthConfig
from vdk.plugin.control_api_auth.auth_config import AuthConfigFolder
from vdk.plugin.control_api_auth.auth_config import InMemAuthConfiguration
from vdk.plugin.control_api_auth.auth_exception import VDKAuthException
from vdk.plugin.control_api_auth.auth_request_values import AuthRequestValues
from vdk.plugin.control_api_auth.login_types import LoginTypes

log = logging.getLogger(__name__)


class AuthenticationCache:
    """
    Class used to keep the fields in the vdk-cred.json file from which
    authorization and authentication tokens can be taken. Used only when
    caching is enabled.
    """

    def __init__(
        self,
        authorization_url=None,
        refresh_token=None,
        access_token=None,
        access_token_expiration_time=0,
        client_id=None,
        client_secret=None,
        api_token=None,
        api_token_authorization_url=None,
        auth_type=None,
        **ignored_kwargs,
    ):
        """
        :param authorization_url:
            Token or Authorization URI used to exchange grant for access token
        :param refresh_token:
            Used in refresh token grant type flow;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-1.5
        :param access_token:
            Cached access token used during requests;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-1.4
        :param access_token_expiration_time:
            When the access token expires
            (epoch seconds -- same as time.time() + "expires_in" value returned
            by the authorization server that generates the access token.)
        :param client_id:
            The client identifier;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-2.3.1
        :param api_token:
            Token used to grant access to the API,
            including exchange for access token.
        :param api_token_authorization_url:
            The URL which exchanges api_token for access token.
        """
        #
        self.authorization_url = authorization_url
        self.refresh_token = refresh_token
        self.access_token = access_token
        self.access_token_expiration_time = access_token_expiration_time
        self.client_id = client_id
        self.client_secret = client_secret
        self.api_token = api_token
        self.api_token_authorization_url = api_token_authorization_url
        self.auth_type = auth_type

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class AuthenticationCacheSerDe:
    """Class used for cache serialization and deserialization."""

    @staticmethod
    def serialize(cache) -> str:
        return json.dumps(cache.__dict__)

    @staticmethod
    def deserialize(content: str) -> AuthenticationCache:
        if content:
            return AuthenticationCache(**json.loads(content))
        else:
            return AuthenticationCache()


class BaseAuth:
    REFRESH_TOKEN_GRANT_TYPE = "refresh_token"  # nosec

    def __init__(
        self,
        conf: Union[AuthConfigFolder, InMemAuthConfiguration] = AuthConfigFolder(),
        vdk_config=AuthConfig(),
    ):
        self._conf = conf
        self.__vdk_config = vdk_config
        self._cache = self.__load_cache()

    def __load_cache(self):
        content = self._conf.read_credentials()
        return AuthenticationCacheSerDe.deserialize(content)

    def __update_cache(self):
        self._conf.save_credentials(AuthenticationCacheSerDe.serialize(self._cache))

    def __exchange_api_for_access_token(self):
        try:
            log.debug(
                f"Refresh API token against {self._cache.api_token_authorization_url} "
            )
            client = OAuth2Session()
            token_response = client.refresh_token(
                self._cache.api_token_authorization_url,
                refresh_token=self._cache.api_token,
            )
            log.debug(
                f"Token response received: "
                f"token_type: {token_response.get('token_type', None)}; "
                f"expires_in: {token_response.get('expires_in', None)}; "
                f"scope: {token_response.get('scope', None)}"
            )
        except Exception as e:
            raise VDKAuthException(
                what="Failed to login",
                why=f"Authorization server at {self._cache.api_token_authorization_url} returned error: {str(e)}",
                consequence="Your credentials are not refreshed and VDK control operations that require authentication "
                "will not work.",
                countermeasure="Check error message and follow instructions in it. Check your network connectivity."
                " Check if you can access the Oauth2 server. Retry to login a few times again."
                " Contact SRE Team operating Oauth2 server for help if nothing else works.",
            ) from e
        # Response per spec in https://tools.ietf.org/html/rfc6750#section-4
        # see also https://www.oauth.com/oauth2-servers/access-tokens/access-token-response
        self.update_access_token(
            token_response[AuthRequestValues.ACCESS_TOKEN_KEY.value]
        )
        self.update_access_token_expiration_time(
            time.time()
            + int(token_response.get(AuthRequestValues.EXPIRATION_TIME_KEY.value, "0"))
        )

    def __exchange_refresh_for_access_token(self):
        basic_auth = HTTPBasicAuth(self._cache.client_id, self._cache.client_secret)
        headers = {
            AuthRequestValues.CONTENT_TYPE_HEADER.value: AuthRequestValues.CONTENT_TYPE_URLENCODED.value,
        }
        data = (
            f"grant_type={self.REFRESH_TOKEN_GRANT_TYPE}&"
            + f"refresh_token={self._cache.refresh_token}"
        )
        log.debug(
            f"Refresh access token against {self._cache.authorization_url} grant_type={self.REFRESH_TOKEN_GRANT_TYPE}"
        )
        response = post(
            self._cache.authorization_url, data=data, headers=headers, auth=basic_auth
        )
        json_data = json.loads(response.text)
        log.debug(
            f"Refresh access token response received: status: {response.status_code}"
        )
        self.update_access_token(json_data[AuthRequestValues.ACCESS_TOKEN_KEY.value])
        self.update_access_token_expiration_time(
            time.time() + int(json_data[AuthRequestValues.EXPIRATION_TIME_KEY.value])
        )

    def read_access_token(self) -> Optional[str]:
        """
        Read access token from _cache or fetch it from Authorization server.
        If not available in _cache it will get it using provided configuration during VDK CLI login to fetch it.
        If it detects that token is about to expire it will try to refresh it.
        :return: the access token or None if it cannot detect any credentials.
        """
        if (
            not self._cache.access_token
            or self._cache.access_token_expiration_time < time.time() + 60
        ):
            log.debug("Acquire access token (it's either expired or missing) ...")
            self.acquire_and_cache_access_token()
        return self._cache.access_token

    def acquire_and_cache_access_token(self):
        """
        Acquires and caches access token
        """
        log.debug(f"Using auth type {self._cache.auth_type} to acquire access token")
        if (
            self._cache.auth_type == LoginTypes.API_TOKEN.value
            and self._configured_api_token()
        ):
            self.__exchange_api_for_access_token()
        elif (
            self._cache.auth_type == LoginTypes.CREDENTIALS.value
            and self._configured_refresh_token()
        ):
            self.__exchange_refresh_for_access_token()
        elif (
            self.__vdk_config.api_token
            and self.__vdk_config.api_token_authorization_url
        ):
            self.update_api_token(self.__vdk_config.api_token)
            self.update_api_token_authorization_url(
                self.__vdk_config.api_token_authorization_url
            )
            self.__exchange_api_for_access_token()
        else:
            log.debug(
                "No authentication mechanism found. Will not cache access token."
                "If Control Service authentication is enabled, API calls will fail."
            )

    def update_api_token_authorization_url(self, api_token_authorization_url):
        """
        Updates and caches authorization URL
        """
        self._cache.api_token_authorization_url = api_token_authorization_url
        self.__update_cache()

    def update_api_token(self, api_token):
        """
        Updates and caches refresh token
        """
        self._cache.api_token = api_token
        self.__update_cache()

    def update_refresh_token(self, refresh_token):
        """
        Updates and caches refresh token
        """
        self._cache.refresh_token = refresh_token
        self.__update_cache()

    def update_client_id(self, client_id):
        """
        Updates and caches refresh token
        """
        self._cache.client_id = client_id
        self.__update_cache()

    def update_client_secret(self, client_secret):
        """
        Updates and caches refresh token
        """
        self._cache.client_secret = client_secret
        self.__update_cache()

    def update_access_token(self, access_token):
        """
        Updates and caches refresh token
        """
        self._cache.access_token = access_token
        self.__update_cache()

    def update_access_token_expiration_time(self, access_token_expiration_time):
        """
        Updates and caches refresh token
        """
        self._cache.access_token_expiration_time = access_token_expiration_time
        self.__update_cache()

    def update_oauth2_authorization_url(self, authorization_url):
        self._cache.authorization_url = authorization_url
        self.__update_cache()

    def update_auth_type(self, auth_type):
        self._cache.auth_type = auth_type
        self.__update_cache()

    def _configured_api_token(self):
        return self._cache.api_token and self._cache.api_token_authorization_url

    def _configured_refresh_token(self):
        return (
            self._cache.refresh_token
            and self._cache.authorization_url
            and self._cache.client_id
            and self._cache.client_secret
        )



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pytest
from pytest_httpserver.pytest_plugin import PluginHTTPServer
from test_core_auth import allow_oauthlib_insecure_transport
from test_core_auth import get_json_response_mock
from vdk.plugin.control_api_auth.auth_exception import VDKInvalidAuthParamError
from vdk.plugin.control_api_auth.authentication import Authentication


def test_api_token_success_authentication(httpserver: PluginHTTPServer):
    allow_oauthlib_insecure_transport()
    httpserver.expect_request("/foo").respond_with_json(get_json_response_mock())

    auth = Authentication(
        token="apitoken",
        authorization_url=httpserver.url_for("/foo"),
        auth_type="api-token",
    )
    auth.authenticate()

    assert auth.read_access_token() == "axczfe12casASDCz"


def test_api_token_no_auth_url():
    auth = Authentication(token="apitoken", auth_type="api-token")

    with pytest.raises(VDKInvalidAuthParamError) as exc_info:
        auth.authenticate()

    raised_exception = exc_info.value
    assert "auth_url was not specified" in raised_exception.message


def test_api_token_no_auth_type_specified(httpserver: PluginHTTPServer):
    httpserver.expect_request("/foo").respond_with_json(get_json_response_mock())
    auth = Authentication(
        token="apitoken", authorization_url=httpserver.url_for("/foo")
    )

    with pytest.raises(VDKInvalidAuthParamError) as exc_info:
        auth.authenticate()
    raised_exception = exc_info.value

    assert "auth_type was not specified" in raised_exception.message



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.plugin.control_api_auth.auth_config import InMemAuthConfiguration
from vdk.plugin.control_api_auth.auth_exception import VDKInvalidAuthParamError
from vdk.plugin.control_api_auth.autorization_code_auth import RedirectAuthentication
from vdk.plugin.control_api_auth.base_auth import BaseAuth
from vdk.plugin.control_api_auth.login_types import LoginTypes


class Authentication:
    """Main class used for authentication."""

    def __init__(
        self,
        username: str = None,
        password: str = None,
        client_id: str = None,
        client_secret: str = None,
        token: str = None,
        authorization_url: str = None,
        auth_discovery_url: str = None,
        auth_type: str = None,
        cache_locally: bool = False,
    ):
        """
        :param username: A user's username in case basic authentication is used.
        :param password: A user's password in case basic authentication is used.
        :param client_id:
            The client identifier;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-2.3.1
        :param client_secret:
            The client identifier;
            See https://datatracker.ietf.org/doc/html/rfc6749#section-2.3.1
        :param token:
            OAuth api token or a refresh token as specified in
            https://datatracker.ietf.org/doc/html/rfc6749#section-1.5 and used
            to obtain access token from the authorization server.
        :param authorization_url:
            The URL which exchanges token for access token.
        :param auth_discovery_url:
            Token or Authorization URI used to exchange grant for access token.
        :param auth_type:
            What type of authentication should be used (e.g., refresh_token,
            basic_auth, etc.).
        :param cache_locally:
            A flag, indicating if credentials should be cached locally (in a
            file).
        """
        self._username = username
        self._password = password
        self._client_id = client_id
        self._client_secret = client_secret
        self._token = token
        self._auth_url = authorization_url
        self._auth_discovery_url = auth_discovery_url
        self._auth_type = auth_type
        # Check if credentials should be cached on the local filesystem
        if cache_locally:
            self._auth = BaseAuth()
        else:
            self._auth = BaseAuth(conf=InMemAuthConfiguration())

    def authenticate(self) -> None:
        if not self._auth_type:
            raise VDKInvalidAuthParamError(
                what="Unable to log in.",
                why="auth_type was not specified.",
                consequence="Subsequent requests to Control Service will not "
                " be authenticated.",
                countermeasure="Specify what type of authentication is to be " "used.",
            )
        if not self._auth_url:
            raise VDKInvalidAuthParamError(
                what="Unable to log in.",
                why="auth_url was not specified.",
                consequence="Authentication is not possible. All subsequent "
                "requests to Control Service will not be authenticated.",
                countermeasure="Provide a valid authorization url.",
            )

        if self._auth_type == LoginTypes.API_TOKEN.value:
            self.__authenticate_with_api_token()
        elif self._auth_type == LoginTypes.CREDENTIALS.value:
            self.__authenticate_with_authorization_code()
        else:
            raise VDKInvalidAuthParamError(
                what="Unexpected authentication type.",
                why=f"Unknown auth_type {self._auth_type} was used.",
                consequence="Authentication is not possible.",
                countermeasure="Provide a valid auth_type.",
            )

    def read_access_token(self):
        """Read access token from cache."""
        return self._auth.read_access_token()

    def __authenticate_with_api_token(self):
        """Authenticate by providing only a OAuth2 API token."""
        self._auth.update_api_token_authorization_url(
            api_token_authorization_url=self._auth_url
        )
        self._auth.update_api_token(api_token=self._token)
        self._auth.update_auth_type(auth_type=LoginTypes.API_TOKEN.value)
        self._auth.acquire_and_cache_access_token()

    # NOTE: Implementation to be added with subsequent PR
    def __authenticate_with_authorization_code(self):
        """
        Authenticate with authorization code as described in
        https://datatracker.ietf.org/doc/html/rfc6749#section-1.3.1
        This type of authentication relies on pkce
        https://datatracker.ietf.org/doc/html/rfc7636
        :return:
        """
        if not self._client_id or not self._auth_discovery_url:
            raise VDKInvalidAuthParamError(
                what="Unable to log in.",
                why=f"Login type {self._auth_type} requires client_id and "
                "auth_discovery_url to be specified.",
                consequence="The authentication operation cannot be complete.",
                countermeasure="Specify client_id and auth_discovery_url and "
                "repeat the login operation.",
            )
        else:
            auth_code_flow = RedirectAuthentication(
                client_id=self._client_id,
                client_secret=self._client_secret,
                oauth2_discovery_url=self._auth_discovery_url,
                oauth2_exchange_url=self._auth_url,
                auth=self._auth,
            )
            auth_code_flow.authentication_process()



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import pathlib
from pathlib import Path
from typing import Optional

from vdk.plugin.control_api_auth.auth_exception import VDKAuthOSError

log = logging.getLogger(__name__)


class AuthConfig:
    @property
    def api_token_authorization_url(self) -> str:
        """
        Location of the API Token OAuth2 provider. Same as login --api-token-authorization-server-url
        This is used as default.
        """
        return os.getenv("VDK_API_TOKEN_AUTHORIZATION_URL", None)

    @property
    def api_token(self) -> str:
        """
        API Token for the OAuth2 provider used in exchange for Access Token
        Same as login --api-token.
        This is used as default.
        """
        return os.getenv("VDK_API_TOKEN", None)

    @property
    def local_config_folder(self) -> str:
        """
        Keeps local configuration of Control CLI like cached tokens, authorization uris, etc.

        Solves following use-cases:

        One scenario where two different users would like to be able to log in on the same machine.
        People sometimes share a development environment as well. Concrete example is two different user
        using the same Jenkins machine

        Also, another use-case is if you want to connect to two different Control Services with different authentication.
        For example, during development, a Demo Control Service works with staging provider but in the
        deployment in production works with Production. So in order to switch you need to re-login every time.
        But now you can just switch tabs (with different environment variables set)
        :return:
        """
        return os.getenv("VDK_BASE_CONFIG_FOLDER", str(Path.home()))


class InMemAuthConfiguration:
    """
    A class providing the equivalent of an in-memory AuthConfigFolder.
    This configuration is available as long as the python process is running,
    and is lost when the process is terminated.
    """

    CREDS_KEY = "credentials"

    def __init__(self):
        self._config = dict()

    def save_credentials(self, content: str) -> None:
        self._config[InMemAuthConfiguration.CREDS_KEY] = content

    def delete_credentials(self) -> None:
        self._config[InMemAuthConfiguration.CREDS_KEY] = ""

    def read_credentials(self) -> Optional[str]:
        return self._config.get(InMemAuthConfiguration.CREDS_KEY)


class AuthConfigFolder:
    """
    A class responsible for managing the configuration files in the vdk configuration
    folder
    """

    # TODO: this looks robust: https://github.com/ActiveState/appdirs
    CONFIG_FOLDER_NAME = ".vdk.internal"
    CREDENTIALS_FILE = "vdk-cred.json"
    CONFIGURATION_FILE = "vdk-configuration.ini"

    def __init__(self, base_dir=AuthConfig().local_config_folder):
        self.vdk_config_folder = os.path.join(
            base_dir, AuthConfigFolder.CONFIG_FOLDER_NAME
        )
        if os.path.isfile(self.vdk_config_folder):
            raise VDKAuthOSError(
                what="Credentials file was not created",
                why=f"There is a file named {self.vdk_config_folder}",
                consequence="User won't be able to access Control Service",
                countermeasure=f"Remove file {self.vdk_config_folder}",
            )
        else:
            if not os.path.exists(self.vdk_config_folder) and not os.path.isdir(
                self.vdk_config_folder
            ):
                try:
                    pathlib.Path(self.vdk_config_folder).mkdir(
                        parents=True, exist_ok=True
                    )
                except Exception as e:
                    raise VDKAuthOSError(
                        what="Configuration folder was not created and user is not logged in",
                        why=f"Cannot create: {self.vdk_config_folder} configuration folder: {str(e)}",
                        consequence="User won't be able to access Control Service",
                        countermeasure=f"Check if the client has write access to: {base_dir}",
                    )

    def save_credentials(self, content):
        try:
            credentials_file_path = self.__get_cred_file()
            log.debug(f"Save vdk credential file: {credentials_file_path} ")
            with open(credentials_file_path, "w+") as temp_file:
                temp_file.write(content)
        except OSError as e:
            raise VDKAuthOSError(
                what="Credentials file was not created",
                why=f"Cannot create credentials file: {str(e)}",
                consequence="User won't be able to access Control Service",
                countermeasure="Check permissions of vdk-cred.json file in .vdk folder inside home directory",
            )

    def delete_credentials(self):
        credentials_file_path = self.__get_cred_file()
        log.debug(f"Delete vdk credential file: {credentials_file_path} ...")
        if os.path.exists(credentials_file_path):
            try:
                os.remove(credentials_file_path)
            except OSError as e:
                raise VDKAuthOSError(
                    what="Credentials file was not deleted",
                    why=f"Cannot delete credentials file in .vdk folder in home directory: {str(e)}",
                    consequence="User is not logged out",
                    countermeasure="Check permissions of vdk-cred.json file in .vdk folder in home directory",
                )

    def read_credentials(self):
        try:
            credentials_file_path = self.__get_cred_file()
            log.debug(f"Reading vdk credential file: {credentials_file_path} ...")
            if os.path.isfile(credentials_file_path):
                with open(credentials_file_path) as temp_file:
                    return temp_file.read()
            return ""
        except OSError as e:
            raise VDKAuthOSError(
                what="Credentials file cannot be accessed.",
                why=f"Most likely not login. Error was: {str(e)}",
                consequence="User won't be able to authenticate against Control Service",
                countermeasure="Make sure you have executed vdk login first. ",
            )

    def __get_cred_file(self):
        return os.path.join(self.vdk_config_folder, AuthConfigFolder.CREDENTIALS_FILE)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pytest
from pytest_httpserver.pytest_plugin import PluginHTTPServer
from test_core_auth import allow_oauthlib_insecure_transport
from test_core_auth import get_json_response_mock
from vdk.plugin.control_api_auth.auth_exception import VDKInvalidAuthParamError
from vdk.plugin.control_api_auth.auth_exception import VDKLoginFailedError
from vdk.plugin.control_api_auth.authentication import Authentication
from vdk.plugin.control_api_auth.autorization_code_auth import LoginHandler
from vdk.plugin.control_api_auth.autorization_code_auth import RedirectAuthentication
from vdk.plugin.control_api_auth.base_auth import BaseAuth
from vdk.plugin.control_api_auth.base_auth import InMemAuthConfiguration


def test_verify_redirect_url():
    allow_oauthlib_insecure_transport()
    in_mem_conf = InMemAuthConfiguration()
    auth = BaseAuth(in_mem_conf)

    auth = RedirectAuthentication(
        client_id="client-id",
        client_secret="client-secret",
        oauth2_discovery_url="http://discovery-url",
        oauth2_exchange_url="http://exchange-url",
        auth=auth,
        redirect_uri="http://127.0.0.1",
        redirect_uri_default_port=9999,
    )
    authorization_url = auth._create_authorization_redirect_url()
    assert (
        authorization_url[0]
        == "http://discovery-url?response_type=code&client_id=client-id&redirect_uri=http%3A%2F%2F127.0.0.1%3A9999&state=requested&prompt=login"
    )
    assert authorization_url[1] == "requested"


def test_login_handler_exceptions(httpserver: PluginHTTPServer):
    allow_oauthlib_insecure_transport()
    httpserver.expect_request("/foo").respond_with_json(get_json_response_mock())
    in_mem_conf = InMemAuthConfiguration()
    auth = BaseAuth(in_mem_conf)

    handler = LoginHandler(
        client_id=None,
        client_secret=None,
        exchange_endpoint=httpserver.url_for("/foo"),
        redirect_uri=None,
        code_verifier=None,
        auth=auth,
    )

    with pytest.raises(VDKLoginFailedError) as exc_info:
        handler.login_with_authorization_code(
            path="http://test-url?code=test-auth-code&client_id=client-id&redirect_uri=http%3A%2F%2F127.0.0.1%3A9999&prompt=login"
        )
    raised_exception = exc_info.value
    assert "Failed to login." in raised_exception.message
    assert "Possibly the request was intercepted." in raised_exception.message

    with pytest.raises(VDKLoginFailedError) as exc_info2:
        handler.login_with_authorization_code(
            path="http://test-url?client_id=client-id&redirect_uri=http%3A%2F%2F127.0.0.1%3A9999&state=requested&prompt=login"
        )
    raised_exception2 = exc_info2.value
    assert "Authentication code is empty" in raised_exception2.message
    assert "The user failed to authenticate properly." in raised_exception2.message


def test_authorization_code_no_secret(httpserver: PluginHTTPServer):
    allow_oauthlib_insecure_transport()
    httpserver.expect_request("/foo").respond_with_json(get_json_response_mock())

    auth = Authentication(
        token="apitoken",
        authorization_url=httpserver.url_for("/foo"),
        auth_type="credentials",
    )
    with pytest.raises(VDKInvalidAuthParamError) as exc_info:
        auth.authenticate()

    raised_exception = exc_info.value
    assert "Unable to log in." in raised_exception.message
    assert "Specify client_id and auth_discovery_url" in raised_exception.message



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0


class VDKAuthException(Exception):
    """
    The VDKAuthException is custom exception type following the coding standard
    for error handling: see the project contributing documentation
    """

    def __init__(self, what, why, consequence, countermeasure):
        banner = "¯\\_(ツ)_/¯"
        self.message = (
            f"{banner}\n"
            f"\nWhat: {what}\n"
            f"Why: {why}\n"
            f"Consequences: {consequence}\n"
            f"Countermeasures: {countermeasure}\n"
        )
        super().__init__(self.message)


class VDKInvalidAuthParamError(VDKAuthException):
    """
    The VDKInvalidAuthParamError is a custom exception type derived from
    the base VDKAuthException type. It is raised when a parameter needed
    for authentication is missing/not provided or otherwise invalid.
    """

    def __init__(self, what, why, consequence, countermeasure):
        super().__init__(
            what=what, why=why, consequence=consequence, countermeasure=countermeasure
        )


class VDKLoginFailedError(VDKAuthException):
    """
    The VDKLoginFailedError is a custom exception type derived from the
    base VDKAuthException type. It is raised when an error occurs while
    going through the actual authentication flow, i.e., when something
    happens while connecting to a third party endpoint.
    """

    def __init__(self, what, why, consequence, countermeasure):
        super().__init__(
            what=what, why=why, consequence=consequence, countermeasure=countermeasure
        )


class VDKAuthOSError(VDKAuthException):
    """
    The VDKAuthOSError is a custom exception type derived from the base
    VDKAuthException type. It is raised when an error in the underlying
    Operating System arises.
    """

    def __init__(self, what, why, consequence, countermeasure):
        super().__init__(
            what=what, why=why, consequence=consequence, countermeasure=countermeasure
        )



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pathlib

import setuptools

"""
Builds a package with the help of setuptools in order for this package to be imported in other projects

"plugin-package-template" is the name of the package which contains the plugin/s.

"plugin-template" is the name of the plugin contained within this package. Note that you can include
more than one plugin in a single package. Also note that the contained plugin may have the same name as the package.

"plugin_template" is the Python file which contains the plugin hooks for the corresponding plugin.
"""

__version__ = "0.1.824443273"

setuptools.setup(
    name="vdk-control-api-auth",
    version=__version__,
    url="https://github.com/vmware/versatile-data-kit",
    description="Versatile Data Kit plugin library provides support for "
    "authentication.",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    install_requires=["requests", "requests_oauthlib"],
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
This script contains test cases, verifying the core authentication components
of the plugin, like authentication cache, configurations, etc.
"""
import json
import os
from time import time
from unittest.mock import patch

from pytest_httpserver.pytest_plugin import PluginHTTPServer
from vdk.plugin.control_api_auth.auth_config import InMemAuthConfiguration
from vdk.plugin.control_api_auth.base_auth import AuthenticationCache
from vdk.plugin.control_api_auth.base_auth import AuthenticationCacheSerDe
from vdk.plugin.control_api_auth.base_auth import BaseAuth


def test_auth_cache_serialization_deserialization():
    cache = AuthenticationCache()
    cache_data = AuthenticationCacheSerDe.serialize(cache)
    serde_cache = AuthenticationCacheSerDe.deserialize(cache_data)
    assert cache == serde_cache


def test_auth_cache_backwards_compatibility_obsolete_keys():
    auth = AuthenticationCache("a", "r", "t")
    data_with_obsolete_keys = auth.__dict__
    data_with_obsolete_keys["obsolete_key"] = "value"
    cache_data = json.dumps(data_with_obsolete_keys)
    cache = AuthenticationCacheSerDe.deserialize(cache_data)
    assert cache.access_token == auth.access_token
    assert cache.refresh_token == auth.refresh_token


def test_auth_cache_backwards_compatibility_missing_keys():
    data_with_missing_keys = dict(refresh_token="value")
    cache_data = json.dumps(data_with_missing_keys)
    cache = AuthenticationCacheSerDe.deserialize(cache_data)
    assert cache.refresh_token == "value"
    assert cache.access_token is None


# The Following tests use the AuthConfigFolder to test configuration storage
# ==========================================================================
def test_auth_updates():
    with patch("vdk.plugin.control_api_auth.auth_config.AuthConfigFolder") as mock_conf:
        # after login cache is populated with auth uri and refresh token
        test_cache = AuthenticationCache()
        mock_conf.read_credentials.return_value = AuthenticationCacheSerDe.serialize(
            test_cache
        )
        auth = BaseAuth(mock_conf)
        auth.update_refresh_token("refresh")
        auth.update_access_token("access")
        auth.update_oauth2_authorization_url("http://uri")

        mock_conf.save_credentials.assert_called_with(
            '{"authorization_url": "http://uri", "refresh_token": "refresh", '
            '"access_token": "access", "access_token_expiration_time": 0, '
            '"client_id": null, "client_secret": null, '
            '"api_token": null, "api_token_authorization_url": null, "auth_type": null}'
        )


def test_auth_acquire_token(httpserver: PluginHTTPServer):
    allow_oauthlib_insecure_transport()
    with patch("vdk.plugin.control_api_auth.auth_config.AuthConfigFolder") as mock_conf:
        httpserver.expect_request("/foo").respond_with_json(get_json_response_mock())

        test_cache = AuthenticationCache(
            api_token_authorization_url=httpserver.url_for("/foo"), api_token="refresh"
        )
        # after login cache is populated with auth uri and refresh token
        mock_conf.read_credentials.return_value = AuthenticationCacheSerDe.serialize(
            test_cache
        )

        auth = BaseAuth(mock_conf)
        auth.update_auth_type("api-token")
        auth.acquire_and_cache_access_token()

        saved_auth = AuthenticationCacheSerDe.deserialize(
            mock_conf.save_credentials.call_args[0][0]
        )
        assert saved_auth.access_token == "axczfe12casASDCz"
        assert saved_auth.access_token_expiration_time > time()


def test_auth_read_access_token_expired(httpserver: PluginHTTPServer):
    allow_oauthlib_insecure_transport()
    with patch("vdk.plugin.control_api_auth.auth_config.AuthConfigFolder") as mock_conf:
        httpserver.expect_request("/foo").respond_with_json(get_json_response_mock())

        test_cache = AuthenticationCache(
            api_token_authorization_url=httpserver.url_for("/foo"),
            api_token="refresh",
            access_token="expired-token",
            access_token_expiration_time=time(),
        )
        # after login cache is populated with auth uri and refresh token
        mock_conf.read_credentials.return_value = AuthenticationCacheSerDe.serialize(
            test_cache
        )

        auth = BaseAuth(mock_conf)
        auth.update_auth_type("api-token")
        auth.read_access_token()

        saved_auth = AuthenticationCacheSerDe.deserialize(
            mock_conf.save_credentials.call_args[0][0]
        )
        # verify it is access token is changed:
        assert saved_auth.access_token == "axczfe12casASDCz"


def test_auth_read_access_token():
    with patch("vdk.plugin.control_api_auth.auth_config.AuthConfigFolder") as mock_conf:
        test_cache = AuthenticationCache(
            "http://foo", "refresh", "access-token", time() + 200
        )
        # after login cache is populated with auth uri and refresh token
        mock_conf.read_credentials.return_value = AuthenticationCacheSerDe.serialize(
            test_cache
        )

        auth = BaseAuth(mock_conf)

        assert auth.read_access_token() == "access-token"


# ============================================================================


# The Following tests use the InMemAuthConfiguration to test configuration
# storage
# ==========================================================================
def test_auth_updates_in_memory():
    # after login cache is populated with auth uri and refresh token
    in_mem_conf = InMemAuthConfiguration()
    auth = BaseAuth(in_mem_conf)
    auth.update_refresh_token("refresh")
    auth.update_access_token("access")
    auth.update_oauth2_authorization_url("http://uri")

    result = '{"authorization_url": "http://uri", "refresh_token": "refresh", "access_token": "access", "access_token_expiration_time": 0, "client_id": null, "client_secret": null, "api_token": null, "api_token_authorization_url": null, "auth_type": null}'

    assert in_mem_conf.read_credentials() == result


def test_auth_acquire_token_in_memory(httpserver: PluginHTTPServer):
    allow_oauthlib_insecure_transport()
    httpserver.expect_request("/foo").respond_with_json(get_json_response_mock())
    in_mem_conf = InMemAuthConfiguration()

    auth = BaseAuth(conf=in_mem_conf)

    auth.update_auth_type("api-token")
    auth.update_api_token_authorization_url(httpserver.url_for("/foo"))
    auth.update_api_token("refresh")

    auth.acquire_and_cache_access_token()

    result = json.loads(in_mem_conf.read_credentials())
    assert result.get("access_token") == "axczfe12casASDCz"
    assert result.get("access_token_expiration_time") > time()


def test_auth_read_access_token_expired_in_memory(httpserver: PluginHTTPServer):
    allow_oauthlib_insecure_transport()
    httpserver.expect_request("/foo").respond_with_json(get_json_response_mock())

    test_cache = AuthenticationCache(
        api_token_authorization_url=httpserver.url_for("/foo"),
        api_token="refresh",
        access_token="expired-token",
        access_token_expiration_time=time(),
    )
    in_mem_conf = InMemAuthConfiguration()
    in_mem_conf.save_credentials(AuthenticationCacheSerDe.serialize(test_cache))

    auth = BaseAuth(in_mem_conf)
    auth.update_auth_type("api-token")
    auth.read_access_token()

    saved_auth = json.loads(in_mem_conf.read_credentials())
    # verify it is access token is changed:
    assert saved_auth.get("access_token") == "axczfe12casASDCz"


def test_auth_read_access_token_in_memory():
    test_cache = AuthenticationCache(
        "http://foo", "refresh", "access-token", time() + 200
    )
    in_mem_conf = InMemAuthConfiguration()
    in_mem_conf.save_credentials(AuthenticationCacheSerDe.serialize(test_cache))

    auth = BaseAuth(in_mem_conf)

    assert auth.read_access_token() == "access-token"


# ============================================================================


def get_json_response_mock():
    json_response_mock = {
        "id_token": "",
        "token_type": "bearer",
        "expires_in": 1799,
        "scope": "csp:support_user",
        "access_token": "axczfe12casASDCz",
        "refresh_token": "refresh",
    }
    return json_response_mock


def allow_oauthlib_insecure_transport():
    os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "true"
